"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from collections import OrderedDict
import decimal

import pandas as pd

from pymbs.log import get_logger

logger = get_logger(__name__)

dec = decimal.Decimal
d0 = dec('0')
d1200 = dec('1200')
cleanup = dec('1E-2')

# TODO: Get the initial rate from the model and get the subsequent rates
# from a FRED lookup
LIBOR = 1.31


class Tranche(object):
    """The Tranche object represents a tranche in the deal.
    It is used to store all pertinent information about the tranche,
    including any cash flows calculated for it.

    The current design of this object provides an experimental 'child tranche'
    option, which allows for the creation of pseudo tranches, to enable a
    true tree-like representation of the cash flow structure and the payment
    waterfall.

    Generally speaking, when reversing a deal, the cash flow structure is
    flattened so that only the tranches disclosed in the Prospectus Supplement
    are included in the model.

    The tree representation with its attending pseudo tranches offeres a truer
    rendition of the model, but may not be necessary and requires a greater
    understanding of structured cash flow models. This functionality *may*
    be deprecated and removed in future releases.
    """

    def __init__(
        self, group, id, upb, coupon, floater_formula, floater_cap,
        floater_floor, prin_type, int_type, notional_with, delay, retail, macr,
        final_payment_date, cusip, schedule,
        dated_date=None, next_payment_date=None
    ):
        self._assumed_price = dec('1.00')
        self._cash_flows = {}
        self._child_tranches = []
        self._coupon = coupon
        self._cusip = cusip
        self._dated_date = dated_date
        self._delay = delay
        self._final_payment_date = final_payment_date
        self._floater_cap = floater_cap
        self._floater_floor = floater_floor
        self._floater_formula = floater_formula
        self._group = group
        self._id = id
        self._int_type = int_type
        self._macr = macr
        self._next_payment_date = next_payment_date
        self._notional_with = notional_with
        self._original_upb = upb
        self._periodic_cf = OrderedDict()
        self._periodic_prepay_scenario = ''
        self._price_at_issuance = dec('1.00')
        self._prin_type = prin_type
        self._pro_rated_ratio = dec('0')
        self._retail = retail
        self._schedule = schedule
        self._strips = []
        self._upb = upb

    @property
    def assumed_price(self):
        return self._assumed_price

    @assumed_price.setter
    def assumed_price(self, value):
        self._assumed_price = value

    @property
    def cash_flows(self):
        return self._cash_flows

    @property
    def child_tranches(self):
        return self._child_tranches

    @property
    def coupon(self):
        return self._coupon

    @coupon.setter
    def coupon(self, value):
        self._coupon = value

    @property
    def cusip(self):
        return self._cusip

    @cusip.setter
    def cusip(self, value):
        self._cusip = value

    @property
    def dated_date(self):
        return self._dated_date

    @dated_date.setter
    def dated_date(self, value):
        self._dated_date = value

    @property
    def delay(self):
        return self._delay

    @delay.setter
    def delay(self, value):
        self._delay = value

    @property
    def final_payment_date(self):
        return self._final_payment_date

    @final_payment_date.setter
    def final_payment_date(self, value):
        self._final_payment_date = value

    @property
    def floater_cap(self):
        return self._floater_cap

    @floater_cap.setter
    def floater_cap(self, value):
        self._floater_cap = value

    @property
    def floater_floor(self):
        return self._floater_floor

    @floater_floor.setter
    def floater_floor(self, value):
        self._floater_floor = value

    @property
    def floater_formula(self):
        return self._floater_formula

    @floater_formula.setter
    def floater_formula(self, value):
        self._floater_formula = value

    @property
    def group(self):
        return self._group

    @group.setter
    def group(self, value):
        self._floater_floor = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._floater_floor = value

    @property
    def int_type(self):
        return self._int_type

    @int_type.setter
    def int_type(self, value):
        self._int_type = value

    @property
    def macr(self):
        return self._macr

    @macr.setter
    def macr(self, value):
        self._macr = value

    @property
    def next_payment_date(self):
        return self._next_payment_date

    @next_payment_date.setter
    def next_payment_date(self, value):
        self._next_payment_date = value

    @property
    def notional_with(self):
        return self._notional_with

    @notional_with.setter
    def notional_with(self, value):
        self._notional_with.clear()
        self._notional_with.extend(value)

    @property
    def original_upb(self):
        return self._original_upb

    @original_upb.setter
    def original_upb(self, value):
        self._original_upb = value

    @property
    def periodic_cf(self):
        return self._periodic_cf

    @periodic_cf.setter
    def periodic_cf(self, value):
        self._periodic_cf = value

    @property
    def periodic_prepay_scenario(self):
        return self._periodic_prepay_scenario

    @periodic_prepay_scenario.setter
    def periodic_prepay_scenario(self, value):
        self._periodic_prepay_scenario = value

    @property
    def price_at_issuance(self):
        return self._price_at_issuance

    @price_at_issuance.setter
    def price_at_issuance(self, value):
        self._price_at_issuance = dec(value)

    @property
    def prin_type(self):
        return self._prin_type

    @prin_type.setter
    def prin_type(self, value):
        self._prin_type = value

    @property
    def pro_rated_ratio(self):
        return self._pro_rated_ratio

    @pro_rated_ratio.setter
    def pro_rated_ratio(self, value):
        self._pro_rated_ratio = dec(value)

    @property
    def retail(self):
        return self._retail

    @retail.setter
    def retail(self, value):
        self._retail = value

    @property
    def schedule(self):
        return self._schedule

    @schedule.setter
    def schedule(self, value):
        self._schedule = value

    @property
    def strips(self):
        """This is a list of Notional tranches, whose interest was stripped
        from this tranche. It is the corollary to the 'notional_with'
        attribute.

        TODO: Need to actually populate this list during the loading of the
        model!
        """
        return self._strips

    @strips.setter
    def strips(self, value):
        self._strips.append(value)

    @property
    def upb(self):
        return self._upb

    @upb.setter
    def upb(self, value):
        self._upb = value

    def new_child_tranche(
        self, group, id, upb, coupon, floater_formula, floater_cap,
        floater_floor, prin_type, int_type, notional_with, delay, retail, macr,
        final_payment_date, cusip, schedule,
        dated_date=None, next_payment_date=None
    ):
        tranche = Tranche(
            group, id, upb, coupon, floater_formula, floater_cap,
            floater_floor, prin_type, int_type, notional_with, delay, retail,
            macr, final_payment_date, cusip, schedule,
            dated_date=None, next_payment_date=None
        )
        self.child_tranches.append(tranche)

        return tranche

    def end_periodic_cf(self):
        self.cash_flows[self.periodic_prepay_scenario].append(
            self.periodic_cf.copy()
        )

    def new_periodic_cf(self, prepay_scenario, period, payment_date):
        if prepay_scenario != self.periodic_prepay_scenario:
            self.periodic_prepay_scenario = prepay_scenario
            self.cash_flows[self.periodic_prepay_scenario] = []
            self.upb = self.original_upb
        self.periodic_cf.clear()
        self._initialize_periodic_flow(period, payment_date)

    def _initialize_periodic_flow(self, period, payment_date):
        self.periodic_cf = OrderedDict().fromkeys(
            [
                'period',
                'payment_date',
                'beginning_balance',
                'interest',
                'accrual',
                'principal',
                'ending_balance'
            ]
        )
        self.periodic_cf['period'] = period
        self.periodic_cf['payment_date'] = payment_date
        self.periodic_cf['beginning_balance'] = self.upb
        self.periodic_cf['interest'] = dec('0')
        self.periodic_cf['accrual'] = dec('0')
        self.periodic_cf['principal'] = dec('0')
        self.periodic_cf['ending_balance'] = self.upb

    def pay_accrue(self):
        self.periodic_cf['accrual'] += self.periodic_cf['interest']
        self.periodic_cf['ending_balance'] += self.periodic_cf['accrual']
        self.upb += self.periodic_cf['accrual']
        self.periodic_cf['interest'] = dec('0')

    def pay_interest(self, collat_net_interest):
        balance = self.upb
        if self.floater_formula:
            coupon = min(
                max(dec(eval(self.floater_formula)), self.floater_floor),
                self.floater_cap
            )
        else:
            coupon = self.coupon
        interest_payment = balance * (coupon / d1200)
        self.periodic_cf['interest'] = interest_payment
        collat_net_interest -= interest_payment

        return collat_net_interest

    def pay_principal(self, principal_payment):
        self.upb -= principal_payment
        if self.upb <= cleanup:
            self.upb = d0
        self.periodic_cf['principal'] += principal_payment
        self.periodic_cf['ending_balance'] -= principal_payment

    def reduce_notional_balance(self):
        pass

    def tabulate_cf(self):
        for prepay_scenario in self.cash_flows:
            cash_flow_table = pd.DataFrame.from_dict(
                self.cash_flows[prepay_scenario]
            )
            self.cash_flows[prepay_scenario] = cash_flow_table


class IndexRate(object):
    """Create and Index rate object for each Benchmark Index used in the deal.

    Traditionally, the most popular index for Mortgage-Backed Securites has
    been 1-Month LIBOR (The London Inter-bank Offered Rate).
    However, in light of recent revelations reagrding LIBOR-fixing
    by market makers, it is being phased-out, in favor of SOFR -
    the Secured Overnight Financing Rate.

    Other Index Rate Benchmarks may be used as well.
    """

    def __init__(self, name, benchmark, fred_ticker, initial_rate):
        self._name = name
        self._benchmark = benchmark
        self._fred_ticker = fred_ticker
        self._initial_rate = initial_rate

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def benchmark(self):
        return self._benchmark

    @benchmark.setter
    def benchmark(self, value):
        self._benchmark = value

    @property
    def fred_ticker(self):
        return self._fred_ticker

    @fred_ticker.setter
    def fred_ticker(self, value):
        self._fred_ticker = value

    @property
    def initial_rate(self):
        return self._initial_rate

    @initial_rate.setter
    def initial_rate(self, value):
        self._initial_rate = value



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from collections import OrderedDict
from datetime import datetime
from dateutil.relativedelta import *
import decimal
import json
import os
from typing import Optional

import pandas as pd

from pymbs import payment_rules  # noqa
from pymbs.config import config
from pymbs.enums import ExitCode, PrepayBenchmark
from pymbs.exceptions import DateError, PrepaymentBenchmarkError
from pymbs.utils import _parse_waterfall, _round_dec

dec = decimal.Decimal
d0 = dec('0')
d1 = dec('1')
d12 = dec('12')
d100 = dec('100')


def _load_assumed_collat():
    if not config.terms_sheet:
        print(f"{config.no_deal}")
        return ExitCode.EX_CONFIG
    else:
        ts = config.terms_sheet
    assumed_collat = pd.DataFrame.from_records(
        ts['assumed_collateral']['data'],
        columns=ts['assumed_collateral']['columns']
    )

    assumed_collat.index = range(1, len(assumed_collat) + 1)

    return assumed_collat


def _load_prepayment_scenarios(series_no, group_id):
    pps_json = os.path.join(
        config.project_dir, f"{series_no}", f"{series_no}_pps.json"
    )
    try:
        with open(pps_json, 'r') as ppsj:
            pps = json.load(ppsj)
    except FileNotFoundError:
        print(f"No Terms Sheet exists for Series {series_no} at {pps_json}")
        return ExitCode.EX_NOINPUT

    for s in pps['prepayment_scenarios']:
        if int(s['group_id']) == group_id:
            return s
            break
    else:
        return None


def _run_collat_cf(group_id, repline_num=0):
    # TODO: Run collateral cash flows for known collateral and securitzed
    # collateral too!
    assumed_collat = _load_assumed_collat()
    group_replines = assumed_collat[assumed_collat['group_id'] == group_id]

    # We're using itertuples() here to loop over the dataframe of the replines
    # belonging to one group. This looping isn't ideal, from a performance
    # perspective, but we anticipate that the given dataframe will always be
    # small enough that there shouldn't be much of an impact.
    if repline_num == 0:
        i = 0
        for repline in group_replines.itertuples():
            if i == 0:
                cash_flows = _run_repline_cf(repline)
                i += 1
            else:
                cf = _run_repline_cf(repline)
                cash_flows.add(cf, fill_value=0)
    else:
        cash_flows = _run_repline_cf(
            group_replines[group_replines['repline'] == repline_num]
        )

    return cash_flows


def _run_repline_cf(repline):
    ts = config.terms_sheet

    series_no = ts['meta']['series_no']
    group_id = repline.group_id
    pps = _load_prepayment_scenarios(series_no, group_id)

    prepayment_benchmark = pps.get('prepayment_benchmark')
    prepayment_speeds = pps.get('speeds')
    # TODO: Need to figure out how to handle prepayment vectors correctly
    prepayment_vector = pps.get('vector')

    if not (prepayment_benchmark and prepayment_speeds) or prepayment_vector:
        raise PrepaymentBenchmarkError(
            'None',
            ('You must provide either a Prepayment Benchmark and a Prepayment '
             'Speed, or a Prepayment Vector.')
        )

    payment_period = ts['meta'].get('payment_period')
    first_payment_date = datetime.strptime(
        ts['meta'].get('first_payment_date'),
        '%Y-%m-%d'
    )

    if not first_payment_date:
        raise DateError(
            'None',
            'You must provide a First Payment Date.'
        )

    if prepayment_benchmark and prepayment_speeds:
        cash_flows = {}
        benchmark = pps['prepayment_benchmark']
        for speed in pps['speeds']:
            sched = pd.DataFrame(
                _amortize_repline(
                    repline.upb,
                    repline.wac,
                    repline.wam,
                    repline.wala,
                    repline.coupon,
                    payment_period,
                    prepayment_benchmark,
                    speed,
                    prepayment_vector,
                    first_payment_date
                )
            )
            sched.index = range(1, len(sched) + 1)
            cash_flows[f"{speed} {benchmark}"] = sched

        return cash_flows


def _amortize_repline(
    orig_upb,
    wac,
    wam,
    wala,
    coupon,
    payment_period=12,
    prepayment_benchmark=None,
    prepayment_speed=None,
    prepayment_vector=[],
    start_date=None
):
    ctx = decimal.getcontext()
    ctx.prec = 18
    ctx.Emax = 12
    ctx.Emin = -10
    decimal.setcontext(ctx)
    payment_period = dec(payment_period)
    prepayment_speed = dec(prepayment_speed)
    # initialize the variables to keep track of
    # the periods and running balances
    p = d1
    beg_balance = orig_upb
    end_balance = orig_upb

    # The Simple Periodic Rate is the wac divided by the payment_period
    spr = (wac / d100) / payment_period

    # The Monthly Passthrough Rate is the coupon divided by the payment_period
    mpr = (coupon / d100) / payment_period

    while end_balance > 0:

        # Recalculate the interest based on the current balance
        # interest = round(((coupon / payment_period) * beg_balance), 2)
        interest = (mpr * beg_balance)

        # Determine Single Monthly Mortality Rate for Current Period
        numerator = spr * ((d1 + spr)**(wam - p + d1))
        denominator = ((d1 + spr)**(wam - p + d1)) - d1

        if denominator > dec('0.00001'):
            pmt = beg_balance * (numerator / denominator)
        else:
            pmt = beg_balance

        smm = _calculate_smm(
            prepayment_benchmark, prepayment_speed, wala, p
        )

        sched_principal = pmt - (spr * beg_balance)

        prepayment = (beg_balance - sched_principal) * smm

        # Ensure additional payment gets adjusted if the loan is being paid off
        prepay_principal = min(prepayment, beg_balance - sched_principal)
        total_principal = sched_principal + prepay_principal
        cash_flow = interest + total_principal
        end_balance = beg_balance - (sched_principal + prepay_principal)

        yield OrderedDict([
            ('period', p),
            ('payment_date', start_date),
            ('beginning_balance', beg_balance),
            ('smm', smm),
            ('scheduled_payment', pmt),
            ('net_interest', interest),
            ('scheduled_principal', sched_principal),
            ('prepayment', prepay_principal),
            ('total_principal', total_principal),
            ('cash_flow', cash_flow),
            ('ending_balance', end_balance)
        ])

        # Increment the counter, balance and date
        p += d1
        start_date += relativedelta(months=1)
        beg_balance = end_balance


def _calculate_smm(prepayment_benchmark, prepayment_speed, wala, period):
    if prepayment_benchmark == PrepayBenchmark.PSA.value:
        benchmark_cpr = dec('0.06')
        seasoned_period = dec('30')

        if (period + wala) < seasoned_period:
            cpr = benchmark_cpr * ((period + wala) / seasoned_period)
        else:
            cpr = benchmark_cpr
        smm = d1 - (d1 - ((prepayment_speed / d100) * cpr))**(d1 / d12)
    elif prepayment_benchmark == PrepayBenchmark.CPR.value:
        pass
    else:
        raise PrepaymentBenchmarkError(
            prepayment_benchmark,
            f"Unknown Prepayment Benchmark: {prepayment_benchmark}"
        )

    return smm


def _calculate_wal(cash_flow, precision, collat_flag: Optional[bool] = False):
    if collat_flag:
        principal = 'total_principal'
    else:
        principal = 'principal'
    cash_flow['period'] = cash_flow['period'].astype(str).transform(dec)
    wal = (sum(
        cash_flow['period'].apply(dec) * cash_flow[f"{principal}"]
    ) / (sum(cash_flow[f"{principal}"]) * d12))

    wal = _round_dec(wal, precision)

    return wal


def _prepare_wals(group, precision, model, _wals):
    _group = model['groups'][group]
    collat_cf = _group['collat_cf']
    for prepay_scenario, cash_flow in collat_cf.items():
        collat_flag = True
        tranche_id = f"Group {group} Collat"
        wal = _calculate_wal(cash_flow, precision, collat_flag)
        _wals['data'].append([group, tranche_id, prepay_scenario, wal])

    if model['groups'][group]['waterfall']:
        collat_flag = False
        for tranche_id in model['groups'][group]['tranches']:
            for prepay_scenario, cash_flow in \
                    _group['tranches'][tranche_id].cash_flows.items():
                    if tranche_id != "SB":
                        wal = _calculate_wal(cash_flow, precision, collat_flag)
                        _wals['data'].append(
                            [group, tranche_id, prepay_scenario, wal]
                        )
    return _wals


def _run_wals(group_id, precision, model):
    _wals = {
        "columns": ["group_id", "tranche_id", "prepay_scenario", "wal"],
        "data": []
    }
    assumed_collat = _load_assumed_collat()
    if group_id < 0:
        groups = assumed_collat['group_id'].unique()
        for group in groups:
            cash_flows = _run_collat_cf(group)
            if model['groups'][group]['waterfall']:
                model['groups'][group]['collat_cf'].update(cash_flows)
                tranches = _get_regular_tranches(model, group)
                _make_payments(model, group, cash_flows, tranches)
            else:
                cash_flows = _run_collat_cf(group)
                model['groups'][group]['collat_cf'].update(cash_flows)
            wals = _prepare_wals(group, precision, model, _wals)
    else:
        cash_flows = _run_collat_cf(group_id)
        if model['groups'][group_id]['waterfall']:
            model['groups'][group_id]['collat_cf'].update(cash_flows)
            tranches = _get_regular_tranches(model, group_id)
            _make_payments(model, group_id, cash_flows, tranches)
        else:
            model['groups'][group_id]['collat_cf'].update(cash_flows)
        wals = _prepare_wals(group_id, precision, model, _wals)

    wal_table = pd.DataFrame.from_records(
        wals['data'],
        columns=wals['columns']
    )

    return wal_table


def _get_regular_tranches(model: dict, group_id: int) -> list:
    tranches = []
    for tranche_id in model['groups'][group_id]['tranches']:
        tranche = model['groups'][group_id]['tranches'][tranche_id]
        if not tranche.macr:
            tranches.append(tranche)

    return tranches


def _make_payments(model, group_id, cash_flows, tranches):
    waterfall = model['groups'][group_id]['waterfall']
    if type(waterfall[0]) is str:
        waterfall = _parse_waterfall(model['groups'][group_id]['waterfall'])
    model['groups'][group_id]['waterfall'] = waterfall

    for prepay_scenario in cash_flows:
        collat_cf = cash_flows[prepay_scenario]
        for payment in collat_cf.itertuples():
            _pay_tranches(
                model, group_id, prepay_scenario, payment, tranches)

    for tranche in tranches:
        tranche.tabulate_cf()


def _pay_tranches(model, group_id, prepay_scenario, payment, tranches):
    interest = payment.net_interest
    for tranche in tranches:
        tranche.new_periodic_cf(
            prepay_scenario,
            payment.period,
            payment.payment_date
        )
        interest = tranche.pay_interest(interest)
    _pay_waterfall(model, group_id, prepay_scenario, payment, tranches)
    for tranche in tranches:
        tranche.end_periodic_cf()


def _pay_waterfall(model, group_id, prepay_scenario, payment, tranches):
    ctx = decimal.getcontext()
    ctx.prec = 18
    ctx.Emax = 12
    ctx.Emin = -10
    decimal.setcontext(ctx)
    waterfall = model['groups'][group_id]['waterfall']
    config.cache.update(
        model=model, prepay_scenario=prepay_scenario,
        group_id=group_id, payment=payment
    )
    for rule in waterfall:
        if rule['tranches']:
            tranches = [
                tranche for tranche in tranches if
                tranche.id in rule['tranches']
            ]
            config.cache.update(tranches=tranches)
        eval(f"payment_rules.{rule['func']}")
